part of 'app_route.dart';

// ignore: prefer_mixin
class AppRouterDelegate extends RouterDelegate<AppPage> with ChangeNotifier, PopNavigatorRouterDelegateMixin {
  static final List<AppPage> _stackPlaceholder = [
    ActivityRoute(location: '/_', builder: (context, args) => const SizedBox.shrink())._createPage()
  ];
  late AppRouteInformationProvider provider;
  late AppRouteInformationParser parser;

  final ApplicationState _app;
  final Map<String, ActivityRoute> _routes;
  final String _initialLocation;
  final OnUnknownRoute _onUnknownRoute;
  final List<AppPage> _stack = [];
  String? _pendingLocation;

  @override
  final GlobalKey<NavigatorState> navigatorKey = GlobalKey();

  @override
  AppPage? get currentConfiguration => _stack.lastOrNull;

  RouteInformation get _entrypoint =>
      RouteInformation(location: WidgetsBinding.instance?.window.defaultRouteName ?? Navigator.defaultRouteName);

  AppRouterDelegate(this._app, Iterable<ActivityRoute> routes, this._initialLocation, this._onUnknownRoute)
      : _routes = Map.fromEntries(routes.map((route) => MapEntry(route.location, route))) {
    provider = AppRouteInformationProvider._(this);
    parser = AppRouteInformationParser._(this);
  }

  @override
  Future<void> setInitialRoutePath(AppPage configuration) {
    final stack = kIsWeb ? _parseRouteStack() : null;
    if (stack == null) return super.setInitialRoutePath(configuration);
    for (final location in stack) {
      final page = configuration.route.location == location
          ? configuration
          : _parseRouteInformationInternal(RouteInformation(location: location));
      if (page != null) _addRoute(page);
    }
    return SynchronousFuture(null);
  }

  @override
  Future<void> setNewRoutePath(AppPage configuration) {
    if (configuration != currentConfiguration) _addRoute(configuration);
    return SynchronousFuture(null);
  }

  @override
  Widget build(BuildContext context) => Navigator(
      key: navigatorKey,
      pages: _stack.isEmpty ? _stackPlaceholder : _stack.toList(growable: false),
      onPopPage: _onPopPage,
      restorationScopeId: 'app_nav');

  @override
  void notifyListeners() {
    super.notifyListeners();
    final title = currentConfiguration?.name;
    if (title != null) _app.title = title;
  }

  Future<bool> didPushRoute(String location) {
    final current = currentConfiguration;
    if (location == _pendingLocation || location == current?.location.toString()) return SynchronousFuture(true);
    _pendingLocation = location;
    final uri = Uri.parse(location);
    final target = _stack.indexWhere((page) => page.location.path == uri.path) + 1;
    if (target == 0) return SynchronousFuture(false);
    if (current == null) return SynchronousFuture(true);

    Future<bool> _popInternal() async {
      var l = _stack.length;
      while (l > target) {
        if (!await navigatorKey.currentState!.maybePop()) break;
        final nl = _stack.length;
        if (nl == l) break;
        l = nl;
      }
      current.updateQueryParameters(_uriParameters(current.arguments, current.route, uri));
      notifyListeners();
      _pendingLocation = null;
      return true;
    }

    return _popInternal();
  }

  Future<T> push<T>(ActivityRoute<T> route, {RouteArgs args = const {}}) async {
    final page = route._createPage(args: args);
    _addRoute(page);
    return page.result;
  }

  bool _onPopPage(Route route, dynamic result) {
    final appPage = route.settings;
    if (route.didPop(result) && appPage is AppPage) {
      _removeRoute(appPage);
      return true;
    }
    return false;
  }

  void _addRoute(AppPage page) {
    _stack.add(page);
    notifyListeners();
  }

  void _removeRoute(AppPage page) {
    _stack.remove(page);
    notifyListeners();
  }

  Future<AppPage> _parseRouteInformation(RouteInformation routeInformation) {
    final page = _parseRouteInformationInternal(routeInformation);
    _pendingLocation = null;
    return SynchronousFuture(page == null ? _unknownRoute(routeInformation) : page);
  }

  AppPage _unknownRoute(RouteInformation routeInformation) {
    final route = _onUnknownRoute(routeInformation);
    var page = route?._createPage(args: {'location': routeInformation.location});
    if (page != null) {
      page._location[0] = Uri.parse(routeInformation.location ?? Navigator.defaultRouteName);
    } else {
      page = currentConfiguration ??
          _parseRouteInformationInternal(const RouteInformation(location: Navigator.defaultRouteName));
      if (page == null) throw StateError('Cannot find route ${routeInformation.location}');
      if (page.location.toString() != routeInformation.location) updateUrl(page.location);
    }
    return page;
  }

  AppPage? _parseRouteInformationInternal(RouteInformation routeInformation) {
    var uri = Uri.parse(routeInformation.location ?? Navigator.defaultRouteName);
    if (uri.path == Navigator.defaultRouteName) uri = uri.replace(path: _initialLocation);
    final route = _routes[uri.path];
    if (route == null) return null;
    final routeArgs = kIsWeb ? _parseRouteInformationState(uri.path) : routeInformation.state as RouteArgs?;
    return route._createPage(args: _uriParameters(routeArgs, route, uri));
  }

  RouteInformation _restoreRouteInformation(AppPage configuration) {
    final uri = configuration._parametrizedUri();
    if (kIsWeb) {
      _restoreRouteInformationState(
          uri.path,
          RouteArgs.from(configuration.arguments)
            ..removeWhere((key, value) => configuration.route.queryParameters.contains(key)));
      _restoreRouteStack(_stack.map((page) => page.route.location).toList(growable: false));
    }
    return RouteInformation(location: uri.toString(), state: configuration.arguments);
  }

  static AppRouterDelegate of(BuildContext context) => Router.of(context).routerDelegate as AppRouterDelegate;

  static RouteArgs _uriParameters(RouteArgs? routeArgs, ActivityRoute route, Uri uri) =>
      (routeArgs != null ? RouteArgs.from(routeArgs) : <String, dynamic>{})
        ..addEntries(uri.queryParameters.entries.where((entry) => route.queryParameters.contains(entry.key)));

  RouteArgs? _parseRouteInformationState(String location) {
    final data = sessionStorage['_route_$location'];
    return data != null ? jsonDecode(data) : null;
  }

  void _restoreRouteInformationState(String location, RouteArgs? args) {
    if (args == null || args.isEmpty) {
      sessionStorage.remove('_route_$location');
    } else {
      sessionStorage['_route_$location'] = jsonEncode(args);
    }
  }

  List<String>? _parseRouteStack() {
    final data = sessionStorage['_route_stack'];
    return data != null ? (jsonDecode(data) as List).cast<String>() : null;
  }

  void _restoreRouteStack(List<String>? locations) {
    if (locations == null || locations.isEmpty) {
      sessionStorage.remove('_route_stack');
    } else {
      sessionStorage['_route_stack'] = jsonEncode(locations);
    }
  }
}
